use cgmath::{Deg, Point2, Point3, Vector2};
use std::time::Instant;
use winit::event::{ElementState, MouseButton, MouseScrollDelta, TouchPhase, VirtualKeyCode};

use super::{ship, Camera, InputStatus, System};
use crate::{
	consts,
	content::Content,
	render::Spriteable,
	render::{Sprite, SpriteTexture},
};

pub struct Projectile {
	pub position: Point2<f32>,
	pub velocity: Vector2<f32>,
	pub sprite: SpriteTexture,
	pub angle: Deg<f32>,
	pub lifetime: f32,
	pub size: f32,
}

impl Projectile {
	pub fn tick(&mut self, t: f32) {
		self.position += self.velocity * t;
		self.lifetime -= t;
	}

	pub fn is_expired(&self) -> bool {
		return self.lifetime < 0.0;
	}
}

pub struct Game {
	pub input: InputStatus,
	pub last_update: Instant,
	pub player: ship::Ship,
	pub test: ship::Ship,
	pub system: System,
	pub camera: Camera,
	paused: bool,
	pub projectiles: Vec<Projectile>,
	pub time_scale: f32,
}

impl Game {
	pub fn new(ct: Content) -> Self {
		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			projectiles: Vec::new(),
			player: ship::Ship::new(
				&ct.ships[0],
				vec![
					ship::ShipOutfit::Gun(ship::ShipGun::new(ct.guns[0].clone(), 1)),
					ship::ShipOutfit::Gun(ship::ShipGun::new(ct.guns[0].clone(), 2)),
					ship::ShipOutfit::Engine(ct.engines[0].clone()),
				],
				(0.0, 0.0).into(),
			),
			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
			},
			system: System::new(&ct.systems[0]),

			paused: false,
			time_scale: 1.0,
			test: ship::Ship::new(&ct.ships[0], vec![], (100.0, 100.0).into()),
		}
	}

	pub fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	pub fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	pub fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	pub fn set_paused(&mut self, pause: bool) {
		if pause {
			self.paused = true;
			self.input.release_all()
		} else {
			self.paused = false;
		}
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		self.projectiles.retain_mut(|p| {
			p.tick(t);
			!p.is_expired()
		});

		// Update player and handle result
		self.player.update_controls(&self.input);
		let mut p = self.player.tick(t);
		if p.projectiles.len() != 0 {
			self.projectiles.append(&mut p.projectiles);
		}

		if self.input.v_scroll != 0.0 {
			self.camera.zoom =
				(self.camera.zoom + self.input.v_scroll).clamp(consts::ZOOM_MIN, consts::ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		self.camera.pos = self.player.physicsbody.pos;

		self.last_update = Instant::now();
	}

	pub fn get_sprites(&self) -> Vec<Sprite> {
		let mut sprites: Vec<Sprite> = Vec::new();

		sprites.append(&mut self.system.get_sprites());
		sprites.push(self.player.get_sprite());
		sprites.push(self.test.get_sprite());

		// Make sure sprites are drawn in the correct order
		// (note the reversed a, b in the comparator)
		//
		// TODO: maybe use a gpu depth buffer instead?
		// I've tried this, but it doesn't seem to work with transparent textures.
		sprites.sort_by(|a, b| b.pos.z.total_cmp(&a.pos.z));

		// Don't waste time sorting these, they should always be on top.
		for p in &self.projectiles {
			sprites.push(Sprite {
				texture: p.sprite.clone(),
				pos: Point3 {
					x: p.position.x,
					y: p.position.y,
					z: 1.0,
				},
				size: p.size,
				angle: p.angle,
				children: None,
			})
		}

		return sprites;
	}
}
